# Sets type hints to act as string so I can reference a class (in a type hint) before it's defined.
from __future__ import annotations

# For autodetecting accessed columns as a test run, and set the trace function back afterwards.
from sys import settrace, gettrace
# For better typehinting
from typing import List, Callable, Tuple, Set, Hashable, Union, Any
# For better type hinting, and detecting uses of Series.__getitem__ specifically.
from pandas import DataFrame, Series, Index
# For autodecteting test result type
from numpy import bool_
# for chekcing if result is number of invalid line
from numbers import Number


class Test:
    """
    A named predicate to enact on the rows of a :class:`pandas.DataFrame` which test the
    validity of one or more of its columns.
    """

    def __init__(self, predicate: Callable[[DataFrame], List[Hashable]], column_index: Index = None, name: str = None,
                 tested_columns: List[str] = None, ignore_columns: List[str] = None, success_threshold=1):
        """
        :class:`Test` class constructor.

        By default, the class will autodetect which columns are being tested as it runs over
        the database. This is fairly crude though, and detects any columns which the test accesses.
        The `tested_columns` and `ignore_columns` parameters can be used to override or modify this behaviour.

        :param predicate: the predicate which will be used to test the rows of the dataframe in this test.

        :param name: a name for the test which will be displayed when running it and can be accessed via the `name`
        property. By default (and if given `None`) this will be set to the name of the predicate function.

        :param tested_columns: setting this parameter to a list of columns will override this behaviour and cause the
        run method to return the given list as the columns tested.

        :param ignore_columns: columns to ignore in tested-columns autodetection. If `tested_columns` is set,
        columns that appear in it and in here will be ignored also.
        """
        self.predicate = predicate
        self.column_index = column_index
        self.success_threshold = success_threshold

        if name is None:
            self.name = self.predicate.__name__
        else:
            self.name = name

        self.tested_columns = set() if tested_columns is None else set(tested_columns)
        self.ignore_columns = set() if ignore_columns is None else set(ignore_columns)

        if column_index is None and (tested_columns is None or len(tested_columns) == 0):
            raise ValueError("Tried to initialize test with no column index and no specified test columns!")

    def test(self, test_target: Any) -> Tuple[Union[bool, List[Hashable]], Set[str]]:
        """
        :param test_target: the row to test
        :returns: a tuple of the boolean test result and the set of columns that were tested.
        """

        # If tested_columns is set then no autodetection of tested columns is needed, just test the row
        # and return the defined tested columns (minus ones specified to ignore).
        if len(self.tested_columns) > 0:
            return self.predicate(test_target), self.tested_columns.difference(self.ignore_columns)

        # Else, the autodetecting tested columns is needed.
        # This is done via initializing an empty accessed_columns
        # set, then running the test with a tracing function which
        # detects accessed columns and adds them to the set

        accessed_columns = set()

        # More info about sys.settrace and the parameters of trace functions can be found
        # at https://explog.in/notes/settrace.html
        def add_accessed_columns(frame, event, arg):
            # Run on *calls* of *__getitem__* with a *Series* object *which is the test row*
            if event == 'call':
                if frame.f_code == Index.__getitem__.__code__:
                    accessed_columns.add(self.column_index[frame.f_locals['key']])
                if frame.f_code == Series.__getitem__.__code__:
                    key = frame.f_locals['key']
                    if key in self.column_index:
                        accessed_columns.add(key)

        original_trace = gettrace()
        settrace(add_accessed_columns)

        # Make sure to eventually set the trace function back.
        try:
            test_result = self.predicate(test_target)
        finally:
            settrace(original_trace)

        return test_result, accessed_columns.difference(self.ignore_columns)

    def run(self, dataframe: DataFrame) -> TestResult:
        result, columns_tested = self.test(dataframe)
        if isinstance(result, bool) or isinstance(result, bool_):
            return BooleanTestResult(self, columns_tested, len(dataframe.index), result)
        elif isinstance(result, Number):
            return NumberTestResult(self, columns_tested, len(dataframe.index), result)
        elif isinstance(result, list):
            return IndexTestResult(self, columns_tested, len(dataframe.index), result, self.success_threshold)
        else:
            raise ValueError(f'Test {self.name}: Invalid test return type: {type(result).__name__} {str(result)}')


class TestResult:
    """
    The results of a :class:`Test` preformed on a dataframe.
    """

    def __init__(self, origin_test: Test, columns_tested: Set[str], num_tested: int,
                 result: Union[bool, Number, List[Hashable]]):
        """
        :param origin_test: the test this is a result of
        :param columns_tested: the columns that were tested on the dataframe (by name)
        :param result: a test result - either a list of invalid row index, or a number of invalid rows
        """
        self.from_test = origin_test
        self.columns_tested = columns_tested
        self.num_tested = num_tested
        self.result = result


class BooleanTestResult(TestResult):
    def __init__(self, origin_test: Test, columns_tested: Set[str], num_tested: int, result: bool):
        super(BooleanTestResult, self).__init__(origin_test, columns_tested, num_tested, result)
        self.success = result


class NumberTestResult(TestResult):
    def __init__(self, origin_test: Test, columns_tested: Set[str], num_tested: int, result: Number):
        super(NumberTestResult, self).__init__(origin_test, columns_tested, num_tested, result)
        self.num_invalid = self.num_tested - result
        self.success = result

    @property
    def num_valid(self):
        """
        Number of dataframe rows that passes the test
        """
        return self.num_tested - self.num_invalid


class IndexTestResult(TestResult):
    def __init__(self, origin_test: Test, columns_tested: Set[str], num_tested: int, result: List[Hashable], success_threshold: float):
        super(IndexTestResult, self).__init__(origin_test, columns_tested, num_tested, result)
        self.num_invalid = len(result)
        self.invalid_row_index = result

        self.success = self.num_valid / self.num_tested >= success_threshold

    @property
    def num_valid(self):
        """
        Number of dataframe rows that passes the test
        """
        return self.num_tested - self.num_invalid

    def get_invalid_rows(self, src_dataframe: DataFrame) -> DataFrame:
        """
        :param src_dataframe: the dataframe the test was ran on
        :return: a dataframe containing each of the rows invalid under this test
        """
        return src_dataframe.iloc[self.invalid_row_index]


# To adjust in HLS
import colorsys
# To generate cmap, convert names to hex & rgb
from matplotlib.colors import cnames, to_rgb, hex2color, LinearSegmentedColormap


def make_autopct(data):
    """
    Generates a function for automatic labeling of pie charts by number and percent values
    :param data: the pie chart data
    :returns: an autolabel function to pass to the "autopct" parameter of pyplot.pie
    """

    def autopct(percent):
        """
        Function for automatic labeling of pie charts by number and percent values
        :param percent: the percent of total a section of a pie chart represents
        :return: a string label, formatted "percent (value)"
        """
        total = sum(data)
        value = int(round(percent * total / 100.0))
        return f'{round(percent, 2)}% ({value})'

    return autopct


# Used for reading config files.
def get_func_from_addr(func_addr: str):
    """
    Returns a function object from its "dotted name",practically, this acts like returning :func:`eval` on
    the input, except it will only work for valid attributes.

    :param func_addr: The dotted name of a funciton, like in an import statement or a line of code:
    e.g module.class.function, module.function, function_name...
    :returns: a function object of the callable the input string refers to.
    """
    func_addr = func_addr.split('.')
    base_attr = __import__(func_addr[0])
    del func_addr[0]
    for attr in func_addr:
        try:
            base_attr = getattr(base_attr, attr)
        except AttributeError as e:
            raise ValueError(f'Tried to parse invalid function name: {".".join(func_addr)}. No {attr} in {base_attr}! '
                             f'Listed in {base_attr}:\n' + "\n".join(dir(base_attr)))

    return base_attr


def to_hex(color):
    # if type(color) == tuple:
    #     return f"#{}"
    # if color.startswith('#'):
    #     return color
    try:
        return cnames[color]
    except KeyError:
        return color


# From https://stackoverflow.com/a/49601444/10913212
def adjust_lightness(color, amount=0.5):
    """
    adjust the given color by multiplying (1-luminosity) by the given amount.
    Input can be matplotlib color string, hex string, or RGB tuple.

    Examples:
    >> lighten_color('g', 0.3)
    >> lighten_color('#F034A3', 0.6)
    >> lighten_color((.3,.55,.1), 0.5)
    """
    color_hex = to_hex(color)
    color_hls = colorsys.rgb_to_hls(*to_rgb(color_hex))
    return colorsys.hls_to_rgb(
        color_hls[0],  # Hue remains
        max(0, min(1, amount * color_hls[1])),  #
        color_hls[2]
    )


# From https://stackoverflow.com/a/50230769/10913212
def nonlinear_cmap(step_colors, step_values, name=None):
    """
    Generates a non-linear matplolib color map. colormap meaning a gradient for color coding values
    in a range (e.g for heat map), and non-linear meaning each color takes over a differently-sized
    chunk of the gradient.

    :param step_colors: an ordered list of colors for each "step" on the gradient
    :param step_values: values on the unit interval matching where each step should be on the graient
    :return: a :class:`matplotlib.color.LinearSegmentedColormap` which can be used as a color map for matplotlib and
    seaborn methods
    """
    color_dict = {'red': (), 'green': (), 'blue': ()}
    for step_color, step_value in zip(step_colors, step_values):
        rgb = hex2color(to_hex(step_color))
        color_dict['red'] = color_dict['red'] + ((step_value, rgb[0], rgb[0]),)
        color_dict['green'] = color_dict['green'] + ((step_value, rgb[1], rgb[1]),)
        color_dict['blue'] = color_dict['blue'] + ((step_value, rgb[2], rgb[2]),)
    return LinearSegmentedColormap(name, color_dict)



# For better type hinting
from typing import Any, List, Union, Iterable
# For dataframe type hinting
from pandas import DataFrame, Series


def from_bool_arr(bool_arr: Iterable[bool]):
    """
    Returns a list of false value indexes from a boolean iterable.

    :param bool_arr: a boolean iterable
    :return: a list of indexes matching false values in the iterable
    """
    return [i for i, cell in enumerate(bool_arr) if not cell]


def in_range_test(left_limit: Union[int, float] = None, right_limit: Union[int, float] = None,
                  left_inclusive=True, right_inclusive=False, preprocess: callable = lambda x: x):
    """
    Creates a generic test function that checks if values for a certain column fall in a given range

    :param left_limit: the lower limit of the range. By default, None, meaning unbounded
    :param right_limit: the upper limit of the range. By default, None, meaning unbounded
    :param left_inclusive: do values equaling the left limit pass the test? By default, false.
    :param right_inclusive: do values equaling the right limit pass the test? By default, false.
    :param preprocess: a function to preprocess values, to ensure comparisons don't raise type errors.
    :return: a test function that checks if values in a dataframe column stay within the specified range
    """
    def test_cell(value):
        return (left_limit is None or left_limit < value) and (right_limit is None or value < right_limit) \
               or left_inclusive and value == left_limit \
               or right_inclusive and value == right_limit

    func = lambda column, df: from_bool_arr(df[column].apply(preprocess).apply(test_cell))
    func.__name__ = f'In range {"[" if left_inclusive else "("}{left_limit}, {right_limit}{"]" if right_inclusive else ")"}'
    return func


def non_equal_test(value: Any):
    """
    Creates a generic test function that checks if values are not equal some given other value.

    :param value: the value to check against
    :return: a generic test function that checks if values are not equal some given other value.
    """
    func = lambda column, df: from_bool_arr((df[column] != value))
    func.__name__ = f'Not {str(value)}'
    return func


def in_list_test(lst: List[Any]):
    """
    Creates a test that checks if values are in a given list

    :param lst: the list to check against
    :return: a generic test function that checks if values in the given column are in the specified list
    """
    func = lambda column, df: from_bool_arr(df[column].apply(lambda x: x in lst))
    func.__name__ = 'In list ' + str(lst)
    return func


def match_test(regex: str):
    """
    :param regex: regular expression to test against
    :return: a generic test function that checks values in the given column match the specified regex
    """
    func = lambda column, df: from_bool_arr(df[column].str.match(regex))
    func.__name__ = f'Match /{regex}/'
    return func


def simple_type_test(data_types: Union[List[type], type]):
    """
    Creates a generic test that checks values are of the specified type.
    Note that this just checks aginst the raw values in the dataframe, which pandas may unexpectedly cast to other types.
    For example, a column of integers with missing values will be automatically converted to floats, to maintain consistency ac
    :param data_types:
    :return:
    """
    data_types = [data_types] if type(data_types) is type else data_types
    type_str = str([data_type.__name__ for data_type in data_types]) if len(data_types) > 1 else data_types[0].__name__

    func = lambda column, df: from_bool_arr(df[column].apply(lambda x: type(x) in data_types))
    func.__name__ = f'Type {type_str}'
    return func


def is_not_null(column: str, dataframe: DataFrame):
    return from_bool_arr(~dataframe[column].isnull())


is_fraction = in_range_test(0, 1, True, True)
is_positive = in_range_test(0, left_inclusive=False)

is_integer = simple_type_test(int)
is_float = simple_type_test(float)
is_str = simple_type_test(str)



from __future__ import annotations

# To parse style files
import json
# Better type hints
from typing import Dict, List


class Style:
    def __init__(self, raw: List[Dict[str, float]] = None):
        if raw is None:
            self.values = [('red', 0),
                           ('orange', 0.25),
                           ('yellow', 0.5),
                           ('blue', 0.75),
                           ('green', 1)]
        else:
            self.values = [(list(item.keys())[0], list(item.values())[0]) for item in raw]

        self.transposed = tuple(map(list, zip(*self.values)))

    @property
    def edges(self):
        return self.values[0][0], self.values[-1][0]

    def colorcode(self, validity_rate: float):
        color_code = None
        for color, step in self.values:
            if validity_rate >= step:
                color_code = color
            else:
                return color_code
        return 'grey' if color_code is None else color_code  # no valid integrity-levels specified or passed all of them


class StyleFile:
    def __init__(self, filename=None):
        if filename is not None:
            with open(filename, 'r') as stylefile:
                parsed = json.load(stylefile)
        else:
            parsed = {}

        if '__DATAFRAME__' in parsed.keys():
            self.dataframe_style = Style(parsed['__DATAFRAME__'])
            del parsed['__DATAFRAME__']
        else:
            self.dataframe_style = Style()

        if '__DEFAULT__' in parsed.keys():
            self.default_style = Style(parsed['__DEFAULT__'])
            del parsed['__DEFAULT__']
        else:
            self.default_style = Style()

        self.styles = {column: Style(palette) for column, palette in parsed.items()}

    def get_column_style(self, column):
        return self.styles.get(column, self.default_style)



# Sets type hints to act as string so I can reference a class (in a type hint) before it's defined.
from __future__ import annotations

# For creating a test for a specific column out of a more generic one
import datetime
# For compressing generic column, row tests to individual column
import operator
# For splitting with delimiter escape in conf parsing
import re
# For patrsing test conf files
from configparser import ConfigParser
# partial for dividing generic tests, reduce for flattening list of invalid index lists
from functools import partial, reduce
# For running pandasgui in the background (so execution is not blocked until user closes it)
from multiprocessing import Process
# For better type hinting
from typing import List, Callable, Hashable, Iterable, Union
# For graphing
from matplotlib import pyplot as plt
# For better type hinting, and detecting uses of Series.__getitem__ specifically.
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Patch
from pandas import DataFrame, Index

# For getting a dataframe in testing (via read_csv) and setting dataframe to not print dimensions (via options
# attribute). from-import not used to make the purpose of these methods more explicit.
import numpy as np
# For  configuring dataframe printing with pandas.options
import pandas
# For heatmaps
import seaborn
# For checking if running in Colab
import sys

import frametests.utils as utils
from frametests.Test import TestResult, Test, IndexTestResult, BooleanTestResult
from frametests.style import StyleFile, Style

# Certain GUIs throw an exception when imported or used in colab.
if 'google.colab' not in sys.modules:
    # For displaying coverage and result graphs
    import pandasgui
    # For graph graphics
    import matplotlib

    matplotlib.use('TkAgg')


class DBTests:
    """
    a testing suite for the data integrity of pandas dataframes.

    this class wraps a pandas dataframe, and can have tests added to to it, simply as
    functions that run on a dataframe row.

    the method :func:`dbtests.run` will iterate over the rows of the database, run the given tests
    and show a coverage report by column — how many of the columns were tested, how many were valid,
    and a sample of invalid rows for each column.
    """

    def __init__(self, df: DataFrame):
        self.dataframe: DataFrame = df
        self.tests: List[Test] = []
        self.columns_tested = set()

    def load_config(self, config_file: str):
        """
        Loads a JSON config file.

        ```
        [testmodule.my_generic_test_func]
        name = mytestbane
        include = col1_to_test,col2_to_test
        exclude = col1_to_ignore,col2_to_ignore
        autodetect = true
        autodetect_ignore = some_cols
        ignore = false
        ```
        any tests specified in brackets like above will be added. all of the settings below it are optional. Generic vs.
        normal tests are detected autpmatically

        for generic tests, include and exclude determine which columns will be tested, and autodetect toggle autodetection.

        for normal tests, include overrides column autodetection and exclude specifies columns to ignore in
        autodetection (or in reading include). "autodetection" value has no effect.

        set ignore to true in order to easily disable tests. name controls test name.
        """
        config = ConfigParser()
        config.read(config_file)

        def get_list(cfg, key: str):
            if key not in cfg.keys():
                return None
            return re.split(r'(?<!\\),', cfg[key])

        for test in config.sections():
            test_cfg = config[test]

            if not test_cfg.getboolean('ignore', False):
                try:
                    test_func = utils.get_func_from_addr(test)
                except ValueError as e:
                    raise ValueError(f'Nonexistent test specified in {config_file}: {test}.')

                argcount = test_func.__code__.co_argcount

                name = test_cfg.get('name', None)
                if argcount == 2:
                    self.add_generic_test(test_func, get_list(test_cfg, 'include'), get_list(test_cfg, 'exclude'),
                                          name, test_cfg.getboolean('autodetect', False),
                                          get_list(test_cfg, 'autodetect_ignore'))
                elif argcount == 1:
                    self.add_test(test_func, name, get_list(test_cfg, 'tested_columns'),
                                  get_list(test_cfg, 'autodetect_ignore'))
                else:
                    raise ValueError(f'Invalid test specified: {test}: function argcount {argcount};'
                                     f'only (row) or (column, row) params are allowed')

    def add_test(self, test_func: Callable[[DataFrame], List[Hashable]], name: str = None,
                 tested_columns: List[str] = None, ignore_columns: List[str] = None):
        """
        Add a test to the Testing Suite.

        By default, the class will autodetect which columns are being tested as it runs over
        the database. This is fairly crude though, and detects any columns which the test accesses.
        The `tested_columns` and `ignore_columns` parameters can be used to override or modify this behaviour.

        :param test_func: the predicate which will be used to test the rows of the dataframe in this test.

        :param name: a name for the test which will be displayed when running it and can be accessed via the `name`
        property. By default (and if given `None`) this will be set to the name of the predicate function.

        :param tested_columns: setting this parameter to a list of columns will override this behaviour and cause the
        run method to return the given list as the columns tested.

        :param ignore_columns: columns to ignore in tested-columns autodetection. If `tested_columns` is set,
        columns that appear in it and in here will be ignored also.
        """
        self.tests.append(Test(test_func, self.dataframe.columns, name, tested_columns, ignore_columns))

    def add_generic_test(self, test_func: Callable[[DataFrame], List[Hashable]], include: Iterable[str] = None,
                         exclude: Iterable[str] = None, name: str = None, column_autodetect: bool = False,
                         ignore_columns: List[str] = None):
        """
        Adds a generic test to a group of columns (or all columns). Instead of as in :func:`add_test`, the
        predicate will not only take a row parameter, but also a column parameter preceding it. Individual
        tests will be added for each of the given columns with the predicate column parameter set to them.

        :param test_func: a predicate will be used to test the rows of the dataframe with respect to a given column.

        :param include: the columns this test should run on. Default is all the columns in the dataframe.

        :param name: a name for the test which will be displayed when running it and can be accessed via the `name`
        property. By default (and if given `None`) this will be set to the name of the predicate function.

        :param column_autodetect: don't set the given column as the tested one, and instead
        autodetect tested columns at runtime. Default false.

        :param ignore_columns: columns to ignore in tested-columns autodetection. Unless column_autodetect is set to
        True, this has no effect
        """
        include = self.dataframe.columns if include is None else include
        exclude = [] if exclude is None else exclude
        for column in (set(include) - set(exclude)):
            tested_cols = None if column_autodetect else [column]
            func_name = test_func.__name__ if name is None else name
            self.add_test(partial(test_func, column), func_name + ' — ' + column, tested_cols, ignore_columns)

    def clear(self):
        """
        Removes all added tests
        """
        self.tests.clear()

    def run(self):
        """
        Runs the given tests over the dataframe and returns a matching :class:`DBTestResults` object
        """
        results = []
        for i, test in enumerate(self.tests):
            print(f'\rTesting {round(i / len(self.tests) * 100):02d}% (#{i + 1}: {test.name})', end='')
            results.append(test.run(self.dataframe))
        print('\rFinished testing')

        return DBTestResults(
            self.dataframe,
            datetime.datetime.now().strftime('%s'),
            results
        )


class ColumnResults:
    """
    The results of tests over a column using one or more multiple predicates.

    This class is used to display results for individual columns, as opposed to thw whole dataframe, both ny writing
    to stdout, displaying summary graphs for the column or individual tests and showing the invalid rows for the column
    in pandasgui.
    """

    def __init__(self, column: str, results: List[TestResult], dataframe: DataFrame, style: Style):
        """
        :param column: the name of the column
        :param results: a list of each :class:`TestResult` result from the tests ran over the columns.
        :param dataframe: the tested dataframe
        """
        self.column = column
        self.results = results
        self.style = style
        self.dataframe = dataframe

        self.style = style
        try:
            self.invalid_row_index = set(reduce(operator.concat, [result.result for result in results]))
        except TypeError:  # Reduce throws a TypeError if it's given an empty list.
            self.invalid_row_index = set()

    @property
    def tested(self):
        return len(self.results) > 0

    @property
    def valid(self):
        """
        Whether the column completely passes each of the tests ran.
        """
        return all(result.success for result in self.results) if len(self.results) > 0 else False

    @property
    def num_tests(self):
        """
        The number of tests ran over the column
        """
        return len(self.results)

    @property
    def num_rows(self):
        return len(self.dataframe.index)

    @property
    def num_invalid(self):
        """
        The number of rows where cells in this column failed at least one test.
        """
        return len(self.invalid_row_index)

    @property
    def num_valid(self):
        """
        The number of rows where cells in this column passed all tests.
        """
        return self.num_rows - self.num_invalid

    def graph_tests_success(self) -> plt.Figure:
        """
        Generates a stacked bar chart showcasing the number of successes (in green)
        and failures (in red) of each test in the column

        Note: In order for the graphs to show, you have to call `pyplot.show()`

        :returns: the graph's pyplot figure
        """
        labels = [result.from_test.name for result in self.results]
        valid_nums = [result.num_valid for result in self.results]
        valid_colors = [self.style.colorcode(valid_num / self.num_rows) for valid_num in valid_nums]
        invalid_nums = [result.num_invalid for result in self.results]
        invalid_colors = [utils.adjust_lightness(valid_color, 0.4) for valid_color in valid_colors]

        fig, axis = plt.subplots()
        axis.bar(labels, valid_nums, label='Valid Rows', color=valid_colors)
        axis.bar(labels, invalid_nums, label='Invalid Rows', bottom=valid_nums, color=invalid_colors)
        axis.legend()
        return fig

    def graph_validity_heatmap(self, binary: bool = None):

        if binary is None:
            binary = any(isinstance(result, BooleanTestResult) for result in self.results)

        test_labels = [result.from_test.name for result in self.results] + [self.column]
        fig, ax = plt.subplots()

        if binary:
            data = [result.success for result in self.results if isinstance(result, IndexTestResult)]
            data.append(all(data))
            data = np.array(data)

            colors = self.style.edges
            color_map = LinearSegmentedColormap.from_list('edges', colors)

            seaborn.heatmap([data],
                            vmin=0, vmax=1,
                            cmap=color_map,
                            cbar=False,
                            linewidth=0.1, linecolor='lightgrey',
                            xticklabels=labels, yticklabels=False)

            legend_handles = [Patch(color=colors[True], label='Valid'),
                              Patch(color=colors[False], label='Invalid')]
            plt.legend(handles=legend_handles)

        else:
            data = np.array(
                [result.num_valid / self.num_rows for result in self.results if isinstance(result, IndexTestResult)])

            step_colors, step_values = self.style.transposed
            color_map = utils.nonlinear_cmap(step_colors, step_values)

            seaborn.heatmap([data],
                            vmin=0, vmax=1,
                            square=True,
                            cmap=color_map,
                            cbar_kws=dict(use_gridspec=False, location="bottom"),
                            annot=True, fmt='.1%',
                            xticklabels=test_labels, yticklabels=False)

            for t in ax.texts:
                t.set_text(t.get_text() + " %")

    def graph_validity(self) -> plt.Figure:
        """
        Generates a pie graph of the column validity as a pyplot figure

        Note: In order for the graphs to show, you have to call `pyplot.show()`

        :returns: the graph's pyplot figure
        """
        fig = plt.figure()
        data = [self.num_valid, self.num_invalid]
        labels = ['Valid', 'Invalid']
        colors = [self.style.values[-1][0], self.style.values[0][0]]

        plt.pie(data, autopct=utils.make_autopct(data), colors=colors)
        fig.legend(labels)
        fig.suptitle(self.column + ' Validity')
        return fig

    def open_invalid_rows(self, index: Union[Index, List[str]] = None, sample_size: int = None):
        """
        Opens the invalid rows at the specified columns in the pandasgui interface.

        :param index: an iterable of the columns to include. This will always include this column.
        :param sample_size: if specified, opens the first n invalid rows.
        """
        index = {self.column} if index is None else set(index).union({self.column})
        failures = [result.get_invalid_rows(self.dataframe)[index] for result in self.results if not result.success and isinstance(result, IndexTestResult)]
        if sample_size is not None:
            failures = [failure.sample(sample_size) for failure in failures]
        pandas_proc = Process(target=pandasgui.show, args=tuple(failures))
        return pandas_proc.start()

    def print(self, columns_to_include=None, column_number=None, print_all_failed=False):
        """
        Prints the results of the column to stdout. Generally, this will generate a title with the column name,
        and print each test, the valid:rows ratio for it and row s where it failed (up to 10)

        :param columns_to_include: columns to include when printing rows of failure.
        By default, only row number and this column are printed.

        :param column_number: specify a column index in the printed title.

        :param print_all_failed: print all invalid rows for each test. By default, prints up to 10.
        """
        prefix = f'Column {column_number}: ' if columns_to_include is not None else ''
        print(f'--- {prefix}{self.column} ---')
        for j, result in enumerate(self.results, 1):
            print(f'Test #{str(j).zfill(2)}: {result.from_test.name}: ', end='')

            if isinstance(result, BooleanTestResult):
                print('Success' if result.success else 'Failure')
            else:
                print(f'{result.num_valid}/{self.num_rows} '
                      f'({round(result.num_valid / self.num_rows * 100, 2)}%).')

                pandas.options.display.show_dimensions = False  # Don't show dimensions when printing rows.
                if print_all_failed:
                    to_print = result.get_invalid_rows(self.dataframe)
                else:
                    to_print = result.get_invalid_rows(self.dataframe).iloc[:10]

                columns_to_include = set() if columns_to_include is None else set(columns_to_include)
                columns_to_include = columns_to_include.union({self.column})
                if not len(to_print.index) == 0:
                    print(to_print[columns_to_include])
                if not print_all_failed and result.num_invalid > 10:
                    print('...')

            print()


class DBTestResults:
    plt = plt

    def __init__(self, dataframe, timestamp, results: List[TestResult]):
        self.dataframe: DataFrame = dataframe
        self.timestamp: int = timestamp
        self.results = results

        self.cols_checked = set().union(*(result.columns_tested for result in results))
        self.invalid_row_index = set().union(
            *(result.result for result in results if isinstance(result, IndexTestResult)))

        self.stylefile = StyleFile()

    @property
    def num_rows(self):
        """Number of rows tested"""
        return len(self.dataframe.index)

    @property
    def num_cols_tested(self):
        """Number of columns tested"""
        return len(self.cols_checked)

    @property
    def num_cols_untested(self):
        return len(self.dataframe.columns) - self.num_cols_tested

    @property
    def num_cols_valid(self):
        return sum(1 for column in self.column_results if column.tested and column.valid)

    @property
    def num_cols_invalid(self):
        return self.num_cols_tested - self.num_cols_valid

    @property
    def num_rows_invalid(self):
        """Number of rows that came as invalid under at least one test (Relevant only for Index Tests)"""
        return len(self.invalid_row_index)

    @property
    def num_rows_valid(self):
        """Number of rows that passes all tests ran (relevant only for Index Tests)"""
        return self.num_rows - self.num_rows_invalid

    @property
    def column_results(self):
        """The list of :class:`ColumnResults` objects for each column in the tested dataframe"""
        return [self.get_column_results(column) for column in self.dataframe.columns]

    def load_styles(self, stylefile_path):
        self.stylefile = StyleFile(stylefile_path)

    def get_column_results(self, column: str) -> ColumnResults:
        """
        :param column: the column to get the results for
        :return: a :class:`ColumnResults` object containing all the results for tests that tested the specified column
        """
        res_list = [result for result in self.results if column in result.columns_tested]
        return ColumnResults(column, res_list, self.dataframe, self.stylefile.get_column_style(column))

    def graph_validity_heatmap(self, binary: bool = None):
        """
        creates and returns a pyplot figure of a 1D heatmap of the columns by validity, color coded by the default
        integrity levels dictionary set for the database.

        :param binary: graph by binary success-failure. If set to false, will use only Index Tests
        and graph rate of validty with a gradient. By default, True if any Boolean Tests were preformed and false otherwise.
        :return: The pyplot figure containing the graph
        """

        if binary is None:
            binary = any(isinstance(result, BooleanTestResult) for result in self.results)

        labels = list(self.dataframe.columns) + ['Dataframe']
        mask = [not self.get_column_results(column).tested for column in self.dataframe.columns]
        mask.append(all(mask))
        fig, ax = plt.subplots()

        if binary:
            data = [result.valid for result in self.column_results]
            data.append(all(data))
            data = np.array(data)

            colors = self.stylefile.dataframe_style.edges
            color_map = LinearSegmentedColormap.from_list('edges', colors)

            seaborn.heatmap([data],
                            vmin=0, vmax=1,
                            cmap=color_map,
                            cbar=False,
                            mask=mask,
                            linewidth=0.1, linecolor='lightgrey',
                            xticklabels=labels, yticklabels=False)

            legend_handles = [Patch(color=colors[True], label='Valid'),
                              Patch(color=colors[False], label='Invalid')]
            plt.legend(handles=legend_handles)

        else:
            data = np.array(
                [result.num_valid / self.num_rows for result in self.column_results]
                + [self.num_rows_valid / self.num_rows]
            )
            step_colors, step_values = self.stylefile.dataframe_style.transposed
            color_map = utils.nonlinear_cmap(step_colors, step_values)
            seaborn.heatmap([data],
                            vmin=0, vmax=1,
                            cmap=color_map,
                            mask=mask,
                            annot=True, fmt='.1%',
                            annot_kws={'rotation': 90},
                            xticklabels=labels, yticklabels=False)

            for t in ax.texts:
                t.set_text(t.get_text() + " %")

        return fig

    def graph_summary(self):
        """
        Adds and returns a pyplot figure containing two pie charts - one for tested vs. untested columns, and one for valid vs. invalid columns.
        """
        colors = [self.stylefile.dataframe_style.values[-1][0], self.stylefile.dataframe_style.values[0][0]]

        fig, (ax1, ax2) = plt.subplots(2)

        tested_data = [self.num_cols_tested, self.num_cols_untested]
        ax1.pie(tested_data, colors=colors, autopct=utils.make_autopct(tested_data))
        ax1.legend(['Tested', 'Untested'])

        valid_data = [self.num_rows_valid, self.num_rows_invalid]
        ax2.pie(valid_data, colors=colors, autopct=utils.make_autopct(valid_data), labels=['Valid', 'Invalid'])
        ax2.legend(["Valid", 'Invalic'])

        return fig

    def print(self, show_valid_cols=False, show_untested=False, stub=False, print_all_failed=False):
        """
        Produces a coverage report of the tests done — how many of the columns were tested, how many were valid,
        and a sample of invalid rows for each column.

        :param show_valid_cols: print result summary for columns that were completely valid. Default false.
        :param show_untested: show columns without tests. Default false.
        :param stub: don't print individual data for each column, just portions tested and valid. Default false.
        :param print_all_failed: print all rows where a test failed. By default only prints up to 10.
        """

        num_cols = len(self.dataframe.columns)

        print(f'Columns Tested: {self.num_cols_tested}/{num_cols} ({round(self.num_cols_tested / num_cols * 100)}%).')
        print(
            f'Columns valid: {self.num_cols_valid}/{self.num_cols_tested} ({round(self.num_cols_valid / self.num_cols_tested * 100, 2)}%).')

        if not stub:  # If stub not set, print details for individual columns.
            print()
            for i, column in enumerate(self.dataframe.columns, 1):
                column_res = self.get_column_results(column)
                # Don't show valid or untested columns unless specified.
                if (column in self.cols_checked or show_untested) \
                        and (column in self.cols_checked and not column_res.valid or show_valid_cols):
                    column_res.print(columns_to_include=[self.dataframe.columns[0]], column_number=i,
                                     print_all_failed=print_all_failed)



from setuptools import setup

setup(
    name='frametests',
    packages=['frametests'],
    version='0.1.1',
    description='A library for testing and analyzing data integrity.',
    author='Atai Ambus',
    license='MIT',
)
